from dataclasses import dataclass

from .manufacturer import Manufacturer


@dataclass
class EEPID:
    """Representation of an EnOcean Equipment Profile (EEP)."""

    def __init__(
        self, rorg: int, func: int, type_: int, manufacturer: Manufacturer | None = None
    ) -> None:
        """Construct an EnOcean Equipment Profile."""
        self.rorg = rorg
        self.func = func
        self.type = type_
        self.manufacturer = manufacturer  # see https://www.enocean.com/wp-content/uploads/application-notes/new_AN514_EnOcean_Link_Profiles.pdf

    @classmethod
    def from_string(cls, eep_string: str) -> "EEPID":
        """Create an EEP instance from a dash-separated string."""
        parts = eep_string.strip().split("-")
        if len(parts) != 3:
            raise ValueError("Wrong format.")
        rorg = int(parts[0], 16)
        func = int(parts[1], 16)
        type_ = int(parts[2], 16)
        return cls(rorg, func, type_)

    def to_string(self) -> str:
        """Return the EEP as a dash-separated string."""
        return f"{self.rorg:02X}-{self.func:02X}-{self.type:02X}"

    def __repr__(self) -> str:
        return f"EEP({self.to_string()}{'.' + str(self.manufacturer) if self.manufacturer is not None else ''})"

    def __hash__(self):
        return hash((self.rorg, self.func, self.type, self.manufacturer))



from .id import EEPID
from .profile import EEP, EEPDataField

EEP_DATABASE: dict[EEPID, EEP] = {
    EEPID.from_string("F6-02-01"): EEP(
        id=EEPID.from_string("F6-02-01"),
        name="Light and Blind Control - Application Style 1",
        datafields=[
            EEPDataField(
                id="R1",
                name="Rocker 1st action",
                offset=0,
                size=3,
                range_min=0,
                range_max=7,
                range_enum={
                    0: "Button A1",
                    1: "Button A0",
                    2: "Button B1",
                    3: "Button B0",
                },
            ),
            EEPDataField(
                id="EB",
                name="Energy bow",
                offset=3,
                size=1,
                range_min=0,
                range_max=1,
                range_enum={0: "released", 1: "pressed"},
            ),
            EEPDataField(
                id="R2",
                name="Rocker 2nd action",
                offset=4,
                size=3,
                range_min=0,
                range_max=7,
                range_enum={
                    0: "Button A1",
                    1: "Button A0",
                    2: "Button B1",
                    3: "Button B0",
                },
            ),
            EEPDataField(
                id="SA",
                name="2nd action",
                offset=7,
                size=1,
                range_min=0,
                range_max=1,
                range_enum={0: "No 2nd action", 1: "2nd action valid"},
            ),
        ],
    ),
}
"""A simple in-memory database of supported EEP profiles. For performance reasons, this is implemented as a dictionary mapping EEPID to EEP."""



"""Asynchronous EnOcean Serial Protocol Version 3 (ESP3) implementation."""

import asyncio
from typing import TYPE_CHECKING, Callable

if TYPE_CHECKING:
    from ..gateway import Gateway

import serial_asyncio_fast as serial_asyncio

from enocean_async.esp3.response import ResponseTelegram

from .packet import SYNC_BYTE, ESP3Packet, ESP3PacketType, crc8


class EnOceanSerialProtocol3(asyncio.Protocol):
    """
    Minimal asynchronous EnOcean Serial Protocol Version 3 (ESP3).
    - Parses ESP3 frames
    - Emits raw ESP3 packets
    """

    def __init__(self, gateway: "Gateway"):
        self.__buffer = bytearray()
        self.__gateway: "Gateway" = gateway

    def connection_made(self, transport: serial_asyncio.SerialTransport):
        self.__gateway.connection_made()

    def data_received(self, data: bytes):
        """Process the internal buffer to extract complete ESP3 packets and emit them."""
        self.__buffer.extend(data)

        while True:
            # find sync byte
            try:
                sync_index = self.__buffer.index(SYNC_BYTE)
            except ValueError:
                self.__buffer.clear()
                return

            # drop garbage before sync
            if sync_index > 0:
                del self.__buffer[:sync_index]

            # need at least sync + header + header CRC
            if len(self.__buffer) < 6:
                return

            # read header
            header = self.__buffer[1:5]
            data_len = (header[0] << 8) | header[1]
            opt_len = header[2]
            packet_type = header[3]

            total_len = 1 + 4 + 1 + data_len + opt_len + 1
            if len(self.__buffer) < total_len:
                return

            # validate header CRC
            if self.__buffer[5] != crc8(header):
                del self.__buffer[:1]
                continue

            # extract data + optional
            data_start = 6
            data_end = data_start + data_len
            opt_end = data_end + opt_len

            data = bytes(self.__buffer[data_start:data_end])
            optional = bytes(self.__buffer[data_end:opt_end])

            # validate data CRC
            if self.__buffer[opt_end] != crc8(data + optional):
                del self.__buffer[:1]
                continue

            # create ESP3Packet and process it
            pkt = ESP3Packet(ESP3PacketType(packet_type), data, optional)
            self.__gateway.process_esp3_packet(pkt)

            # Remove processed bytes
            del self.__buffer[:total_len]

    def connection_lost(self, exception: Exception | None):
        self.__gateway.connection_lost(exception)

    def eof_received(self, exception: Exception | None):
        pass



"""EnOcean address classes for Home Assistant integration.

This module provides the EnOceanAddress class for handling EnOcean four byte (32 bit) addresses,
including parsing, validation, and conversion between integer and string formats.

For more information on EnOcean addressing, see
  - https://www.enocean-alliance.org/wp-content/uploads/2021/03/EURID-v1.2.pdf
  - https://www.enocean.com/de/faq-knowledge-base/what-is-difference-between-base-id-and-chip-id/
"""

type SenderAddress = EURID | BaseAddress
"""Addresses that can be used as the sender of a telegram. This includes both EURIDs (device addresses) and Base IDs, hence ranges from 00:00:00:00 to FF:FF:FF:80."""


class Address:
    """Implementation of the EnOcean four byte (32 bit) addresses to identify devices.

    # A note about sending addresses
    Each EnOcean device has a unique ID (called *Chip ID*) which it can use
    for sending telegrams. Alternatively, EnOcean gateways can also use a range
    of 128 consecutive addresses for sending, starting at the so-called
    _Base ID_ of the gateway. A gateway's *Base ID* is a four byte address in
    the range FF:80:00:00 to FF:FF:FF:80. The Base ID is a predefined address,
    which can be changed by the user only a few times (at most 10 times for the
    TCM310 chip). The allowed addresses for sending a telegram are thus the
    following 129 addresses:

    - Chip ID (= device ID),
    - Base ID,
    - Base ID + 1,
    - Base ID + 2,
    - ...
    - Base ID + 126, and
    - Base ID + 127.

    All other addresses must not be used for sending (and will be rejected by
    official EnOcean modules). This is meant as a basic security feature. Have a
    look at the EnOcean [knowledge base](https://www.enocean.com/de/faq-knowledge-base/what-is-difference-between-base-id-and-chip-id/) for the official explanation of the differences between chip ID and base IDs.

    Base IDs are always in the range FF:80:00:00 to FF:FF:FF:80.
    """

    def __init__(self, from_value: int | str) -> None:
        """Initialize the EnOceanID from an integer or string."""
        numeric_id = -1
        if isinstance(from_value, str):
            numeric_id = Address.from_string(from_value).to_number()
        if isinstance(from_value, int):
            numeric_id = from_value
        if not isinstance(numeric_id, int):
            raise TypeError(
                "ID must be an integer or a hex string that can be converted to an integer."
            )
        if numeric_id < 0:
            raise ValueError("ID out of bounds (must be at least 0).")
        if numeric_id > 0xFFFFFFFF:
            raise ValueError(
                "ID out of bounds (must be smaller than 0xFFFFFFFF = 4294967295)."
            )
        self.__address = numeric_id

    @classmethod
    def from_number(cls, id: int) -> "Address":
        """Create an EnOceanID instance from an integer."""
        return cls(id)

    @classmethod
    def from_string(cls, id_string: str) -> "Address":
        """Create an EnOceanID instance from a colon-separated string."""
        if not id_string:
            raise ValueError("from_string called with undefined argument")
        parts = id_string.strip().split(":")
        if len(parts) != 4:
            raise ValueError("Wrong format.")
        hex_string = "".join(part.zfill(2) for part in parts)
        return cls(int(hex_string, 16))

    @classmethod
    def from_bytelist(cls, byte_list: list[int]) -> "Address":
        """Create an EnOceanID instance from a list of 4 bytes."""
        if len(byte_list) != 4:
            raise ValueError("Byte list must have exactly 4 elements.")
        numeric_id = (
            (byte_list[0] << 24)
            | (byte_list[1] << 16)
            | (byte_list[2] << 8)
            | byte_list[3]
        )
        return cls(numeric_id)

    @classmethod
    def broadcast(cls) -> "Address":
        """Return the broadcast ID (FF:FF:FF:FF)."""
        return cls(0xFFFFFFFF)

    @classmethod
    def validate_string(cls, id_string: str) -> bool:
        """Check that the supplied string is a valid EnOcean address."""
        parts = id_string.strip().split(":")

        if len(parts) != 4:
            return False

        hex_string = "".join(part.zfill(2) for part in parts)
        try:
            int(hex_string, 16)
        except ValueError:
            return False

        return True

    def to_number(self) -> int:
        """Return the EnOcean address as integer."""
        return self.__address

    def to_string(self) -> str:
        """Return the EnOcean address as colon-separated hex string."""
        s = f"{self.__address:08X}"
        return f"{s[0:2]}:{s[2:4]}:{s[4:6]}:{s[6:8]}"

    def to_json(self) -> str:
        """Return the EnOcean address as JSON string."""
        return self.to_string()

    def to_bytelist(self) -> list[int]:
        """Return the EnOcean address as list of bytes."""
        return [
            (self.__address >> 24) & 0xFF,
            (self.__address >> 16) & 0xFF,
            (self.__address >> 8) & 0xFF,
            self.__address & 0xFF,
        ]

    def is_eurid(self) -> bool:
        """Check if the address is a EURID (EnOcean Unique Radio Identifier)."""
        return 0x00000000 <= self.__address <= 0xFF7FFFFF

    def is_base_address(self) -> bool:
        """Check if the address is a Base address."""
        return 0xFF800000 <= self.__address <= 0xFFFFFF80

    def is_broadcast(self) -> bool:
        """Check if the address is the broadcast address."""
        return self.__address == 0xFFFFFFFF

    def __str__(self) -> str:
        """Return the EnOcean address as string."""
        return self.to_string()

    def __hash__(self):
        return self.__address

    def __eq__(self, other):
        return self.to_number() == other.to_number()

    def __repr__(self) -> str:
        return f"EnOceanAddress({self.to_string()})"


class EURID(Address):
    """Representation of an EnOcean device address (EnOcean Unique Radio Identifier / EURID).

    Device addresses are in the range 00:00:00:00 to FF:7F:FF:FF.
    """

    def __init__(self, from_value: int | str) -> None:
        """Initialize the EnOcean device address from an integer or string."""
        numeric_address = -1
        if isinstance(from_value, str):
            numeric_address = Address.from_string(from_value).to_number()
        if isinstance(from_value, int):
            numeric_address = from_value
        if not isinstance(numeric_address, int):
            raise TypeError(
                "Address must be an integer or a hex string that can be converted to an integer."
            )
        if not (0x00000000 <= numeric_address <= 0xFF7FFFFF):
            raise ValueError(
                f"Device address must be in the range 00:00:00:00 to FF:7F:FF:FF, but is {numeric_address:08X}."
            )
        super().__init__(numeric_address)

    def __repr__(self) -> str:
        return f"EURID({self.to_string()})"


class BaseAddress(Address):
    """Representation of an EnOcean base address.

    Base addresses are in the range FF:80:00:00 to FF:FF:FF:80.
    """

    def __init__(self, from_value: int | str) -> None:
        """Initialize the EnOcean base address from an integer or string."""
        numeric_address = -1
        if isinstance(from_value, str):
            numeric_address = Address.from_string(from_value).to_number()
        if isinstance(from_value, int):
            numeric_address = from_value
        if not isinstance(numeric_address, int):
            raise TypeError(
                "Address must be an integer or a hex string that can be converted to an integer."
            )
        if not (0xFF800000 <= numeric_address <= 0xFFFFFF80):
            raise ValueError(
                "Base address must be in the range FF:80:00:00 to FF:FF:FF:80."
            )
        super().__init__(numeric_address)

    def __repr__(self) -> str:
        return f"EnOceanBaseAddress({self.to_string()})"


class BroadcastAddress(Address):
    """Representation of the EnOcean broadcast address (FF:FF:FF:FF)."""

    def __init__(self) -> None:
        """Initialize the EnOcean broadcast address."""
        super().__init__(0xFFFFFFFF)

    def __repr__(self) -> str:
        return "EnOceanBroadcastAddress(FF:FF:FF:FF)"



from ..erp1.telegram import ERP1Telegram
from .id import EEPID
from .message import EEPMessage
from .profile import EEP


class EEPHandler:
    """An EEP handler is responsible for encoding and decoding messages for a specific EEP."""

    def __init__(self, eep: EEP):
        self.__eep = eep

    def decode(self, telegram: ERP1Telegram) -> EEPMessage:
        """Convert an ERP1Telegram into an EEPMessage."""

        msg = EEPMessage(sender=telegram.sender, eepid=self.__eep.id, rssi=0, values={})

        # iterate over the data fields defined in the EEP and extract values from the telegram
        for field in self.__eep.datafields:
            value = None
            # if the field has an enumeration, convert the raw value to its corresponding string representation
            if field.range_enum is not None:
                raw_value = telegram.bitstring_raw_value(
                    offset=field.offset, size=field.size
                )
                value = field.range_enum.get(raw_value, f"Unknown({raw_value})")
            elif (
                field.scale_min is not None
                and field.scale_max is not None
                and field.range_min is not None
                and field.range_max is not None
            ):
                value = telegram.bitstring_scaled_value(
                    offset=field.offset,
                    size=field.size,
                    range_min=field.range_min,
                    range_max=field.range_max,
                    scale_min=field.scale_min,
                    scale_max=field.scale_max,
                )
            else:
                value = telegram.bitstring_raw_value(
                    offset=field.offset, size=field.size
                )

            msg.values[field.id] = value

        return msg

    def encode(self, message: EEPMessage) -> ERP1Telegram:
        """Convert an EEPMessage into an ERP1Telegram."""
        # Not all EEPs will support encoding, so this can be left unimplemented if desired.
        pass

    def __call__(self, telegram: ERP1Telegram) -> EEPMessage:
        """Allow decoder instances to be called like functions."""
        return self.decode(telegram)



from dataclasses import dataclass
from enum import IntEnum

from ..erp1.address import BaseAddress
from .packet import ESP3Packet, ESP3PacketType


class CommonCommandCode(IntEnum):
    """Common command codes for ESP3 packets."""

    CO_RD_VERSION = 3
    """Read the device version information"""

    CO_WR_IDBASE = 7
    """Set base ID"""

    CO_RD_IDBASE = 8
    """ Read ID range base address"""


@dataclass
class CommonCommandTelegram:
    """Common Command Telegram for ESP3 packets."""

    common_command_code: int
    common_command_data: bytes | None = None
    optional_data: bytes = None

    @classmethod
    def CO_RD_VERSION(cls) -> "CommonCommandTelegram":
        """Create a Common Command Telegram to read the device version information."""
        return cls(common_command_code=CommonCommandCode.CO_RD_VERSION)

    @classmethod
    def CO_RD_IDBASE(cls) -> "CommonCommandTelegram":
        """Create a Common Command Telegram to read the base ID."""
        return cls(common_command_code=CommonCommandCode.CO_RD_IDBASE)

    @classmethod
    def CO_WR_IDBASE(cls, id_base: BaseAddress) -> "CommonCommandTelegram":
        """Create a Common Command Telegram to set the base ID."""

        id_base_bytes = id_base.to_bytelist()
        return cls(
            common_command_code=CommonCommandCode.CO_WR_IDBASE,
            common_command_data=id_base_bytes,
        )

    def __post_init__(self):
        if self.optional_data is None:
            self.optional_data = b""

    def to_esp3_packet(self) -> ESP3Packet:
        data_size = (
            1 if self.common_command_data is None else len(self.common_command_data) + 1
        )
        data = bytearray(data_size)
        data[0] = self.common_command_code

        if self.common_command_data is not None:
            data[1:] = self.common_command_data

        return ESP3Packet(
            packet_type=ESP3PacketType.COMMON_COMMAND,
            data=bytes(data),
            optional=self.optional_data,
        )



import asyncio
from dataclasses import dataclass
import logging
import time
from typing import Callable, Optional

import serial_asyncio_fast as serial_asyncio

from .eep.db import EEP_DATABASE
from .eep.handler import EEPHandler
from .eep.id import EEPID
from .eep.message import EEPMessage
from .erp1.address import EURID, BaseAddress, SenderAddress
from .erp1.telegram import RORG, ERP1Telegram
from .erp1.ute import (
    EEPTeachInResponseMessageExpectation,
    UTEMessage,
    UTEQueryRequestType,
    UTEResponseType,
)
from .esp3.common_command import CommonCommandTelegram
from .esp3.packet import ESP3Packet, ESP3PacketType
from .esp3.protocol import EnOceanSerialProtocol3
from .esp3.response import ResponseCode, ResponseTelegram
from .version.id import VersionIdentifier
from .version.info import VersionInfo

type RSSI = int


@dataclass
class SendResult:
    response: Optional[ResponseTelegram]
    duration_ms: Optional[float]


# callback types
type ESP3Callback = Callable[[ESP3Packet], None]
type ERP1Callback = Callable[[ERP1Telegram], None]
type EEPMessageCallback = Callable[[EEPMessage], None]
type UTECallback = Callable[[UTEMessage], None]
type ResponseCallback = Callable[[ResponseTelegram], None]
type NewDeviceCallback = Callable[[SenderAddress], None]
type ParsingFailedCallback = Callable[[str], None]


@dataclass
class CallbackWithFilter:
    callback: Callable
    sender_filter: SenderAddress | None = None


@dataclass
class ERP1CallbackWithFilter(CallbackWithFilter):
    callback: ERP1Callback


@dataclass
class EEPCallbackWithFilter(CallbackWithFilter):
    callback: EEPMessageCallback


class BaseIDChangeError(Exception):
    pass


class Gateway:
    """EnOcean gateway that connects to a serial port and processes incoming ESP3 packets."""

    def __init__(self, port: str, baudrate: int = 57600):
        """Create an instance of an EnOcean gateway that connects to the supplied port at supplied baudrate (optional) and processes incoming ESP3 packets."""

        # serial connection, transport and protocol parameters
        self.__port: str = port
        self.__baudrate: int = baudrate
        self.__transport: serial_asyncio.SerialTransport | None = None
        self.__protocol: EnOceanSerialProtocol3 | None = None

        # cached information about the connected module (to avoid unnecessary requests for information that doesn't change)
        self.__version_info: VersionInfo | None = None
        self.__base_id_remaining_write_cycles: int | None = None
        self.__base_id: BaseAddress | None = None

        # device and EEP management
        self.__known_devices: dict[EURID | BaseAddress, EEPID] = {}
        self.__detected_devices: list[EURID | BaseAddress] = []
        self.__eep_handlers: dict[EEPID, EEPHandler] = {}

        # callbacks
        self.__esp3_receive_callbacks: list[ESP3Callback] = []
        self.__erp1_receive_callbacks: list[ERP1CallbackWithFilter] = []
        self.__ute_receive_callbacks: list[UTECallback] = []
        self.__eep_receive_callbacks: list[EEPCallbackWithFilter] = []
        self.__parsing_failed_callbacks: list[ParsingFailedCallback] = []
        self.__response_callbacks: list[ResponseCallback] = []

        self.__new_device_callbacks: list[NewDeviceCallback] = []

        self.__esp3_send_callbacks: list[ESP3Callback] = []

        # send handling
        self.__send_lock: asyncio.Lock = asyncio.Lock()
        self.__send_future: asyncio.Future | None = None

        # learning
        self.__is_learning: bool = False
        self.__learning_timeout_task: asyncio.Task | None = None
        self.__allow_teach_out: bool = False

        # logging
        self._logger = logging.getLogger(__name__)

    # ------------------------------------------------------------------
    # callback registration
    # ------------------------------------------------------------------
    def add_esp3_received_callback(self, cb: ESP3Callback):
        """Add a callback that will be called for every received ESP3 packet.

        This is a low-level callback that will be called for every ESP3 packet as they are received from the serial port, before any parsing or processing. This can be useful for debugging or for implementing custom processing of ESP3 packets that is not covered by the built-in functionality of the Gateway class."""
        self.__esp3_receive_callbacks.append(cb)

    def add_esp3_send_callback(self, cb: ESP3Callback):
        """Add a callback that will be called for every ESP3 packet that is sent to the EnOcean module.

        This can be useful for debugging or for implementing custom logging of sent packets."""
        self.__esp3_send_callbacks.append(cb)

    def add_new_device_callback(self, cb: NewDeviceCallback):
        """Add a callback that will be called for every newly detected sender address (EURID or Base ID) from incoming ERP1 telegrams.

        This can be useful for implementing custom handling of new devices."""
        self.__new_device_callbacks.append(cb)

    def add_erp1_received_callback(
        self, cb: ERP1Callback, sender_filter: SenderAddress | None = None
    ):
        """Add a callback that will be called for every received ERP1 telegram. If sender_filter is provided, the callback will only be called for telegrams that have a sender address matching the filter.

        This is a semi-high-level callback that will be called for every received ERP1 telegram after parsing and basic processing, but before any EEP-specific decoding. This can be useful for handling ERP1 telegrams in a custom way, for example by implementing custom decoding for specific RORGs or by handling telegrams from unknown devices."""
        self.__erp1_receive_callbacks.append(ERP1CallbackWithFilter(cb, sender_filter))

    def add_eep_message_received_callback(
        self, cb: EEPMessageCallback, sender_filter: SenderAddress | None = None
    ):
        """Add a callback that will be called for every received ERP1 telegram that could successfully be decoded as an EEP message. If sender_filter is provided, the callback will only be called for messages that have a sender address matching the filter.

        This is a high-level callback that will be called for every received EEP message after parsing, basic processing, and EEP-specific decoding. Prerequisite for this callback to be called for a message are:
        - the sender address of the message is known (by adding it to this gateway as known-device along with its EEPID), and
        - there is an EEPHandler capable of handling the EEPID of the sender device."""
        self.__eep_receive_callbacks.append(EEPCallbackWithFilter(cb, sender_filter))

    def add_ute_received_callback(self, cb: UTECallback):
        self.__ute_receive_callbacks.append(cb)

    def add_parsing_failed_callback(self, cb: ParsingFailedCallback):
        self.__parsing_failed_callbacks.append(cb)

    def add_response_callback(self, cb: ResponseCallback):
        self.__response_callbacks.append(cb)

    # ------------------------------------------------------------------
    # start and stop
    # ------------------------------------------------------------------
    async def start(self) -> None:
        """Open the serial connection to the EnOcean module and start processing incoming packets."""
        loop = asyncio.get_running_loop()
        try:
            (
                self.__transport,
                self.__protocol,
            ) = await serial_asyncio.create_serial_connection(
                loop,
                lambda: EnOceanSerialProtocol3(self),
                self.__port,
                baudrate=self.__baudrate,
            )

            self._logger.info(
                f"Successfully connected to EnOcean module on {self.__port} at baudrate {self.__baudrate}"
            )
        except Exception as e:
            self._logger.error(
                f"Failed to connect to EnOcean module on {self.__port} at baudrate {self.__baudrate}: {e}"
            )
            raise ConnectionError(
                f"Failed to connect to EnOcean module on {self.__port} at baudrate {self.__baudrate}: {e}"
            )

    def stop(self) -> None:
        """Close the serial connection to the EnOcean module."""
        if self.__transport is not None:
            self.__transport.close()
            self.__transport = None
            self._logger.info(
                f"Serial connection to EnOcean module on {self.__port} closed"
            )

    def start_learning(
        self, timeout_seconds: int = 60, allow_teach_out: bool = False
    ) -> None:
        """Start learning mode."""
        self.__is_learning = True
        self.__allow_teach_out = allow_teach_out

        self._logger.info(
            f"Learning mode started. Will automatically stop after {timeout_seconds} seconds."
        )
        if self.__learning_timeout_task is not None:
            self.__learning_timeout_task.cancel()
        self.__learning_timeout_task = asyncio.create_task(
            self.__learning_timeout(timeout_seconds)
        )

    def stop_learning(self) -> None:
        """Stop learning mode."""
        self.__is_learning = False
        self._logger.info("Learning mode stopped.")
        if self.__learning_timeout_task is not None:
            self.__learning_timeout_task.cancel()
            self.__learning_timeout_task = None

    async def __learning_timeout(self, timeout_seconds: int):
        try:
            await asyncio.sleep(timeout_seconds)
            if self.__is_learning:
                self.stop_learning()
        except asyncio.CancelledError:
            pass

    # ------------------------------------------------------------------
    # sending commands and receiving responses
    # ------------------------------------------------------------------
    async def send_esp3_packet(self, packet: ESP3Packet) -> SendResult:
        """Send an ESP3 packet to the EnOcean module and wait up to 500ms for a response (as per ESP3 specification).

        This method is thread-safe and can be called from multiple coroutines concurrently; the send operations will be serialized using an internal lock, and each call will wait for its corresponding response before allowing the next send operation to proceed. The method returns a SendResult object containing the received response (if any) and the duration in milliseconds between sending the request and receiving the response.
        """

        async with self.__send_lock:
            self.__send_future = asyncio.get_running_loop().create_future()

            try:
                # emit to the send callbacks; we do this before sending the packet (WHY?)
                self.__emit(self.__esp3_send_callbacks, packet)
                self._logger.debug(
                    f"Sending ESP3 packet: {packet}. Waiting for response..."
                )

                # start a timer before sending the packet, so that we can measure the time it takes to receive the response after sending the packet
                start = time.perf_counter()

                # send the frame
                self.__transport.write(packet.to_bytes())
                response: ResponseTelegram | None = await asyncio.wait_for(
                    self.__send_future, timeout=0.5
                )

                # stop the timer and calculate duration
                end = time.perf_counter()

                self._logger.debug(
                    f"Received response to sent packet: {response}. Duration: {(end - start) * 1000:.2f} ms"
                )

                return SendResult(response, (end - start) * 1000)

            finally:
                self.__send_future = None

    def connection_made(self) -> None:
        pass

    def connection_lost(self, exc: Exception | None) -> None:
        self.__transport = None

    # ------------------------------------------------------------------
    # device registry
    # ------------------------------------------------------------------
    def add_device(self, address: EURID | BaseAddress, eepid: EEPID) -> None:
        """Register a device with its sender address (EURID or Base ID) and its EEPID.

        This allows the gateway to recognize incoming messages from this device and decode them according to the registered EEP (if a handler for that EEP is found).
        """
        self.__known_devices[address] = eepid
        self._logger.info(f"Added device with address {address} and EEPID {eepid}")

        # get the EEP handler for this EEPID
        if not eepid in self.__eep_handlers:
            # try to load
            if eepid not in EEP_DATABASE:
                self._logger.warning(
                    f"EEP {eepid} is not supported. Messages from device {address} will not be decoded."
                )
                return
            else:
                self.__eep_handlers[eepid] = EEPHandler(EEP_DATABASE[eepid])
                self._logger.info(f"Loaded EEP handler for EEPID {eepid}")
        else:
            self._logger.debug(f"EEP handler for EEPID {eepid} already loaded.")

    def remove_device(self, address: EURID | BaseAddress) -> None:
        """Deregister a device by its sender address (EURID or Base ID). This removes the device from the registry of known devices, so that incoming messages from this address will no longer be recognized as coming from a known device and will not be decoded as EEP messages."""
        if address in self.__known_devices:
            del self.__known_devices[address]
            self._logger.info(f"Removed device with address {address}")
        else:
            self._logger.warning(
                f"Tried to remove device with address {address}, but it was not found in the registry of known devices."
            )

    # ------------------------------------------------------------------
    # Gateway properties and methods
    # ------------------------------------------------------------------
    @property
    async def base_id(self) -> BaseAddress | None:
        """Get the base ID of the connected EnOcean module."""
        if self.__base_id is not None:
            return self.__base_id

        if self.__transport is None:
            raise ConnectionError("Not connected to EnOcean module")

        # Send GET ID base id request
        cmd = CommonCommandTelegram.CO_RD_IDBASE()
        result: SendResult = await self.send_esp3_packet(cmd.to_esp3_packet())
        response = result.response

        if (
            response is None
            or response.return_code != ResponseCode.OK
            or len(response.response_data) < 4
        ):
            return None

        self.__base_id = BaseAddress.from_bytelist(response.response_data[:4])

        if len(response.optional_data) >= 1:
            self.__base_id_remaining_write_cycles = response.optional_data[0]

        return self.__base_id

    async def change_base_id(
        self, new_base_id: BaseAddress, safety_flag: int = 0
    ) -> BaseAddress | None:
        """Change the base ID of the connected EnOcean module. Returns True if successful, False otherwise.

        To prevent accidental changes, this function requires a safety flag to be provided. The safety_flag must be set to 0x7B, otherwise the function will return False without sending the command."""

        if safety_flag != 0x7B:
            raise ValueError(
                "Invalid safety flag. To change the base ID, you must provide a safety flag of 0x7B to confirm that you understand the consequences of this action."
            )

        if self.__transport is None:
            raise ConnectionError("Not connected to EnOcean module")

        base_id_before_change = (
            await self.base_id
        )  # store previous base ID for error message in case the change failed

        if new_base_id == base_id_before_change:
            raise ValueError("New base ID is the same as the current base ID")

        # send WR ID base id request
        cmd = CommonCommandTelegram.CO_WR_IDBASE(new_base_id)
        send_result = await self.send_esp3_packet(cmd.to_esp3_packet())
        response = send_result.response

        # check response for errors; if we got a response, but it indicates an error, we can be pretty sure that the base ID change failed, so we can raise an exception with the error message
        if response is not None and response.return_code != ResponseCode.OK:
            match response.return_code:
                case ResponseCode.NOT_SUPPORTED:
                    raise BaseIDChangeError(
                        "Base ID change is not supported by this module"
                    )
                case ResponseCode.BASEID_OUT_OF_RANGE:
                    raise BaseIDChangeError(
                        "Base ID change failed: provided base ID is out of allowed range (must be in range FF:80:00:00 to FF:FF:FF:80)"
                    )
                case ResponseCode.BASEID_MAX_REACHED:
                    raise BaseIDChangeError(
                        "Base ID change failed: maximum number of allowed base ID changes has been reached"
                    )
                case _:
                    raise BaseIDChangeError(
                        f"Base ID change failed with error code: {response.return_code.name} ({response.return_code.value})"
                    )

        # now either we got a successful response, or no response at all (timeout). In both cases, we should check if the base ID was actually changed by reading it again, because the module might have accepted the command but failed to send a response.
        self.__base_id = None  # reset cached base ID to force re-fetching it
        self.__base_id_remaining_write_cycles = (
            None  # reset cached remaining write cycles as well
        )
        reported_base_id = await self.base_id
        if reported_base_id == new_base_id:
            return reported_base_id
        elif reported_base_id == base_id_before_change:
            raise BaseIDChangeError(
                "Base ID change failed: after sending the command, the base ID of the module is still the same as before"
            )
        else:
            raise BaseIDChangeError(
                f"Base ID change failed: after sending the command, the module reports a different base ID ({reported_base_id}) than the one we tried to set ({new_base_id}) and the one that was set before ({base_id_before_change}), which is very unexpected and likely indicates a communication error or a bug in the module's firmware."
            )

    @property
    async def version_info(self) -> VersionInfo | None:
        """Get the version information of the connected EnOcean module."""
        if self.__version_info is not None:
            return self.__version_info

        if self.__transport is None:
            raise ConnectionError("Not connected to EnOcean module")

        # Send GET VERSION request
        cmd = CommonCommandTelegram.CO_RD_VERSION()
        send_result = await self.send_esp3_packet(cmd.to_esp3_packet())
        response = send_result.response

        if (
            response is None
            or response.return_code != ResponseCode.OK
            or len(response.response_data) < 32
        ):
            return None

        self.__version_info = VersionInfo(
            app_version=VersionIdentifier(
                main=response.response_data[0],
                beta=response.response_data[1],
                alpha=response.response_data[2],
                build=response.response_data[3],
            ),
            api_version=VersionIdentifier(
                main=response.response_data[4],
                beta=response.response_data[5],
                alpha=response.response_data[6],
                build=response.response_data[7],
            ),
            eurid=EURID.from_bytelist(response.response_data[8:12]),
            device_version=response.response_data[12],
            app_description=response.response_data[16:32]
            .decode("ascii")
            .rstrip("\x00"),
        )

        return self.__version_info

    @property
    async def base_id_remaining_write_cycles(self) -> int | None:
        """Get the remaining write cycles for the base ID of the connected EnOcean module."""
        if self.__base_id_remaining_write_cycles is not None:
            return self.__base_id_remaining_write_cycles

        await (
            self.base_id
        )  # base_id() will fetch the remaining write cycles as optional data
        return self.__base_id_remaining_write_cycles

    @property
    async def eurid(self) -> EURID | None:
        """Get the EURID of the connected EnOcean module."""
        return (await self.version_info).eurid

    # ------------------------------------------------------------------
    # Internal packet processing
    # ------------------------------------------------------------------
    def process_esp3_packet(self, packet: ESP3Packet):
        """Process a received ESP3 packet. This includes emitting the raw packet to registered callbacks and further processing based on packet type."""
        self.__emit(self.__esp3_receive_callbacks, packet)

        self._logger.debug(f"Received ESP3 packet: {packet}")

        # handle packet based on type; currently we only process RESPONSE and RADIO_ERP1 packets, other types are ignored
        if packet.packet_type == ESP3PacketType.RESPONSE:
            response: ResponseTelegram
            try:
                response = ResponseTelegram.from_esp3_packet(packet)
            except Exception as e:
                return
            self.__process_response(response)
            return

        if packet.packet_type != ESP3PacketType.RADIO_ERP1:
            return

        # parse to ERP1 telegram; if parsing fails, ignore the packet and return
        erp1: ERP1Telegram
        try:
            erp1 = ERP1Telegram.from_esp3(packet)
        except Exception as e:
            return

        self.__process_erp1_telegram(erp1)

    def __emit(self, callbacks: list[Callable], obj):
        """Emit an object to all registered callbacks of the given type."""
        loop = asyncio.get_running_loop()
        for cb in callbacks:
            loop.call_soon(cb, obj)

    def __emit_with_sender_filter(
        self, callbacks: list[CallbackWithFilter], sender: SenderAddress, obj
    ):
        """Emit an object to all registered callbacks of the given type that have a sender filter matching the sender address."""
        loop = asyncio.get_running_loop()
        for cb in callbacks:
            if cb.sender_filter is None or cb.sender_filter == sender:
                loop.call_soon(cb.callback, obj)

    def __is_sender_known(self, sender: SenderAddress) -> bool:
        """Check if the sender address is known (i.e. if we have an EEP ID for it)."""
        return (
            sender in self.__known_devices.keys() or sender in self.__detected_devices
        )

    def __process_response(self, response: ResponseTelegram):
        """Process a received RESPONSE packet. If we are currently awaiting a response, try to parse it and store it for the send() method to retrieve."""
        self.__emit(self.__response_callbacks, response)
        self._logger.debug(f"Processing received RESPONSE packet: {response}")

        if self.__send_future and not self.__send_future.done():
            self.__send_future.set_result(response)

    def __process_erp1_telegram(self, erp1: ERP1Telegram):
        """Process a received ERP1 telegram. This includes emitting it to registered callbacks and further processing based on RORG and learning bit."""
        # emit the raw telegram
        self.__emit_with_sender_filter(self.__erp1_receive_callbacks, erp1.sender, erp1)
        self._logger.debug(f"ESP3 packet successfully decoded to ERP1 telegram: {erp1}")

        # check if sender is known; if not, emit to new device callbacks and add to detected devices list
        if not self.__is_sender_known(erp1.sender):
            self.__detected_devices.append(erp1.sender)
            self.__emit(self.__new_device_callbacks, erp1.sender)
            self._logger.info(f"New device detected with sender address: {erp1.sender}")

        # if it's a UTE telegram, try to parse to UTE message; if parsing fails, ignore the packet and return;
        if erp1.rorg == RORG.RORG_UTE:
            try:
                ute_message = UTEMessage.from_erp1(erp1)
            except Exception as e:
                return

            self.__handle_ute_message(ute_message)
            return

        # handle 1BS teach-in telegrams; for now, we just ignore them (NOT IMPLEMENTED)
        if erp1.rorg == RORG.RORG_1BS and erp1.is_learning_telegram:
            self.__handle_1bs_teach_in_telegram(erp1)
            return

        # handle 4BS teach-in telegrams; for now, we just ignore them (NOT IMPLEMENTED)
        if erp1.rorg == RORG.RORG_4BS and erp1.is_learning_telegram:
            self.__handle_4bs_teach_in_telegram(erp1)
            return

        # if sender is not known, we cannot decode to EEP message, so we return
        if not erp1.sender in self.__known_devices:
            msg = f"Failed to decode ERP1 telegram to EEP message: sender {erp1.sender} is unknown."
            self._logger.debug(msg)
            self.__emit(self.__parsing_failed_callbacks, msg)
            return

        # if we have no EEP handler for the EEP ID of the sender, we cannot decode to EEP message, so we return
        if not self.__known_devices[erp1.sender] in self.__eep_handlers:
            self._logger.debug(
                f"Failed to decode ERP1 telegram to EEP message: No EEP handler for {self.__known_devices[erp1.sender]}."
            )
            self.__emit(
                self.__parsing_failed_callbacks,
                f"Failed to decode ERP1 telegram to EEP message: No EEP handler for {self.__known_devices[erp1.sender]}.",
            )
            return

        # try eep-specific decoding; if it fails, ignore the packet and return
        try:
            eep_message = self.__eep_handlers[self.__known_devices[erp1.sender]](erp1)
            self.__emit_with_sender_filter(
                self.__eep_receive_callbacks, erp1.sender, eep_message
            )
            self._logger.debug(
                f"ERP1 telegram successfully decoded to EEP message: {eep_message}"
            )
        except Exception as e:
            self._logger.debug(f"Failed to decode ERP1 telegram to EEP message: {e}")
            self.__emit(
                self.__parsing_failed_callbacks,
                f"Failed to decode ERP1 telegram to EEP message: {e}",
            )
            return

    def __handle_ute_message(self, ute_message: UTEMessage):
        self.__emit(self.__ute_receive_callbacks, ute_message)

        # if we are not currently in learning mode, we ignore all UTE messages, because they are only relevant during learning mode
        if not self.__is_learning:
            return

        # ignore response messages, we only want to process teach-in query messages during learning mode
        if isinstance(ute_message.request_type, UTEResponseType):
            return

        request_type = ute_message.request_type
        response_expected = (
            ute_message.teach_in_response_message_expectation
            == EEPTeachInResponseMessageExpectation.RESPONSE_EXPECTED
        )

        match request_type:
            case UTEQueryRequestType.TEACH_IN:
                print(f"Received UTE teach-in query message: {ute_message}")

            case UTEQueryRequestType.TEACH_IN_DELETION:
                print(f"Received UTE teach-in deletion query message: {ute_message}")

            case UTEQueryRequestType.TEACH_IN_OR_DELETION_OF_TEACH_IN:
                print(
                    f"Received UTE teach-in or deletion of teach-in query message: {ute_message}"
                )

    def __handle_1bs_teach_in_telegram(self, erp1: ERP1Telegram):
        pass

    def __handle_4bs_teach_in_telegram(self, erp1: ERP1Telegram):
        pass

    def __handle_eep_message(self, msg: EEPMessage):
        for cb in self.callbacks:
            cb(msg)



from dataclasses import dataclass


@dataclass
class VersionIdentifier:
    main: int = 0
    beta: int = 0
    alpha: int = 0
    build: int = 0

    @property
    def version_string(self) -> str:
        return f"{self.main}.{self.beta}.{self.alpha}{f'b{self.build}' if self.build else ''}"



from dataclasses import dataclass
from enum import IntEnum
from typing import Optional

from enocean_async.eep.manufacturer import Manufacturer
from enocean_async.erp1.errors import ERP1ParseError
from enocean_async.esp3.packet import ESP3Packet, ESP3PacketType

from ..eep.id import EEPID
from ..erp1.rorg import RORG
from ..erp1.telegram import ERP1Telegram


class UTEQueryRequestType(IntEnum):
    TEACH_IN = 0
    TEACH_IN_DELETION = 1
    TEACH_IN_OR_DELETION_OF_TEACH_IN = 2
    NOT_USED = 3


class UTEResponseType(IntEnum):
    NOT_ACCEPTED_GENERAL_REASON = 0
    ACCEPTED_TEACH_IN = 1
    ACCEPTED_DELETION_OF_TEACH_IN = 2
    NOT_ACCEPTED_EEP_NOT_SUPPORTED = 3


class CommunicationDuringEEPOperation(IntEnum):
    UNIDIRECTIONAL = 0
    BIDIRECTIONAL = 1


class CommandIdentifier(IntEnum):
    TEACH_IN_QUERY = 0
    TEACH_IN_RESPONSE = 1


class EEPTeachInResponseMessageExpectation(IntEnum):
    RESPONSE_EXPECTED = 0
    NO_TEACH_IN_RESPONSE = 1


@dataclass
class UTEMessage:
    communication_during_eep_operation: CommunicationDuringEEPOperation
    teach_in_response_message_expectation: Optional[
        EEPTeachInResponseMessageExpectation
    ]
    request_type: UTEQueryRequestType | UTEResponseType
    command: CommandIdentifier
    number_of_channels_to_be_taught_in: int
    manufacturer: Manufacturer
    eep: EEPID

    @classmethod
    def from_erp1(cls, telegram: ERP1Telegram) -> "UTEMessage":
        if telegram.rorg != RORG.RORG_UTE:
            raise ValueError(f"Invalid RORG for UTE: {telegram.rorg}")

        if len(telegram.telegram_data) < 6:
            raise ValueError(
                f"Invalid data length for UTE: {len(telegram.telegram_data)} bytes but expected 6 bytes"
            )

        communication_during_eep_operation = CommunicationDuringEEPOperation(
            telegram.bitstring_raw_value(0, 1)
        )

        command = CommandIdentifier(telegram.bitstring_raw_value(4, 4))

        if command == CommandIdentifier.TEACH_IN_QUERY:
            teach_in_response_message_expectation = (
                EEPTeachInResponseMessageExpectation(telegram.bitstring_raw_value(1, 1))
            )

        request_type_value = telegram.bitstring_raw_value(2, 2)
        request_type = (
            UTEQueryRequestType(request_type_value)
            if command == CommandIdentifier.TEACH_IN_QUERY
            else UTEResponseType(request_type_value)
        )

        number_of_channels_to_be_taught_in = telegram.data_byte(5)

        manufacturer_id_lsb = telegram.data_byte(4)
        manufacturer_id_msb = telegram.data_byte(3) & 0b00000111
        manufacturer_id = (manufacturer_id_msb << 8) | manufacturer_id_lsb

        manufacturer: Manufacturer
        try:
            manufacturer = Manufacturer(manufacturer_id)
        except ValueError:
            manufacturer = Manufacturer.Reserved

        eep = EEPID(
            type_=telegram.data_byte(2),
            func=telegram.data_byte(1),
            rorg=telegram.data_byte(0),
        )

        return cls(
            communication_during_eep_operation=communication_during_eep_operation,
            teach_in_response_message_expectation=teach_in_response_message_expectation,
            request_type=request_type,
            command=command,
            number_of_channels_to_be_taught_in=number_of_channels_to_be_taught_in,
            manufacturer=manufacturer,
            eep=eep,
        )

    @classmethod
    def response_for_query(
        cls, query: UTEMessage, response_type: UTEResponseType
    ) -> UTEMessage:
        if query.command != CommandIdentifier.TEACH_IN_QUERY:
            raise ValueError("Could not create response for non-query.")

        return UTEMessage(
            communication_during_eep_operation=query.communication_during_eep_operation,
            teach_in_response_message_expectation=None,
            request_type=response_type,
            command=CommandIdentifier.TEACH_IN_RESPONSE,
            number_of_channels_to_be_taught_in=query.number_of_channels_to_be_taught_in,
            manufacturer=query.manufacturer,
            eep=query.eep,
        )



from dataclasses import dataclass

from ..erp1.address import EURID
from .id import VersionIdentifier


@dataclass
class VersionInfo:
    app_version: VersionIdentifier
    api_version: VersionIdentifier
    eurid: EURID
    device_version: int
    app_description: str



from dataclasses import dataclass

from enocean_async.eep.id import EEPID


@dataclass
class EEPDataField:
    """An EEP data field represents a single data point within an EEP, such as a sensor value or a control command."""

    id: str
    """Unique identifier for the data field, typically a short string like 'R1', 'POS' etc."""

    name: str
    """Human-readable name for the data field, providing a descriptive label."""

    offset: int
    """Bit offset of the data field within the EEP."""

    size: int
    """Size of the data field in bits."""

    range_min: int | None = None
    """Minimum value of the data field's range."""

    range_max: int | None = None
    """Maximum value of the data field's range."""

    scale_min: float | None = None
    """Minimum value of the data field's scaled value."""

    scale_max: float | None = None
    """Maximum value of the data field's scaled value."""

    range_enum: dict[int, str] | None = None
    """Enumeration of possible values for the data field, if applicable."""

    unit: str | None = None
    """Unit of the data field, such as '°C' for temperature or '%' for humidity."""


@dataclass
class EEP:
    """Base class for EEP data."""

    id: EEPID
    """Unique identifier for the EEP."""

    name: str
    """Human-readable name/description for the EEP."""

    datafields: list[EEPDataField]
    """List of data fields within the EEP."""



from dataclasses import dataclass
from typing import Any, Dict

from ..eep.id import EEPID
from ..erp1.address import EURID, BaseAddress


@dataclass
class EEPMessage:
    """An EEP message represents a message according to a specific EEP profile in a a dictionary of values extracted from the message according to the EEP profile's data fields. It should not be generated manually but using a proper EEPHandler.

    For convenience, it can include the sender's address, the EEPID of the message, and the RSSI (signal strength).
    """

    sender: EURID | BaseAddress | None
    """The sender's address, which can be a EURID or a BaseAddress. This is optional and can be None if the sender is unknown or not relevant."""

    eepid: EEPID | None = None
    """The EEPID of the message. This is optional and can be None if the EEPID is unknown or not relevant."""

    rssi: int | None = None
    """The RSSI (signal strength) of the message. This is optional and can be None if the RSSI is unknown or not relevant."""

    values: Dict[str, Any] = None
    """A dictionary of values extracted from the message according to the EEP profile's data fields. The keys are the data field IDs (e.g., 'R1', 'POS'), and the values are the corresponding values extracted from the message."""

    def __repr__(self) -> str:
        return (
            f"<EEPMessage {self.eepid} from {self.sender.to_string()}: {self.values}>"
        )



from enum import IntEnum


class Manufacturer(IntEnum):
    Reserved = 0x000
    PEHA = 0x001
    THERMOKON = 0x002
    SERVODAN = 0x003
    ECHOFLEX_SOLUTIONS = 0x004
    OMNIO_AG = 0x005
    AWAG_ELEKTROTECHNIK_AG = 0x005
    HARDMEIER_ELECTRONICS = 0x006
    REGULVAR_INC = 0x007
    AD_HOC_ELECTRONICS = 0x008
    DISTECH_CONTROLS = 0x009
    KIEBACK_AND_PETER = 0x00A
    ENOCEAN_GMBH = 0x00B
    PROBARE = 0x00C
    VICOS_GMBH = 0x00C
    ELTAKO = 0x00D
    LEVITON = 0x00E
    HONEYWELL = 0x00F
    SPARTAN_PERIPHERAL_DEVICES = 0x010
    SIEMENS = 0x011
    T_MAC = 0x012
    RELIABLE_CONTROLS_CORPORATION = 0x013
    ELSNER_ELEKTRONIK_GMBH = 0x014
    DIEHL_CONTROLS = 0x015
    BSC_COMPUTER = 0x016
    S_AND_S_REGELTECHNIK_GMBH = 0x017
    ZENO_CONTROLS = 0x018
    MASCO_CORPORATION = 0x018
    INTESIS_SOFTWARE_SL = 0x019
    VIESSMANN = 0x01A
    LUTUO_TECHNOLOGY = 0x01B
    CAN2GO = 0x01C
    SAUTER = 0x01D
    BOOT_UP = 0x01E
    OSRAM_SYLVANIA = 0x01F
    UNOTECH = 0x020
    DELTA_CONTROLS_INC = 0x21
    UNITRONIC_AG = 0x022
    NANOSENSE = 0x023
    THE_S4_GROUP = 0x024
    MSR_SOLUTIONS = 0x025
    VEISSMANN_HAUSATOMATION_GMBH = 0x025
    GE = 0x26
    MAICO = 0x027
    RUSKIN_COMPANY = 0x28
    MAGNUM_ENERGY_SOLUTIONS = 0x29
    KMC_CONTROLS = 0x02A
    ECOLOGIX_CONTROLS = 0x02B
    TRIO_2_SYS = 0x2C
    AFRISO_EURO_INDEX = 0x02D
    WALDMANN_GMBH = 0x02E
    NEC_PLATFORMS_LTD = 0x030
    ITEC_CORPORATION = 0x031
    SIMICX_CO_LTD = 0x32
    PERMUNDO_GMBH = 0x33
    EUROTRONIC_TECHNOLOGY_GMBH = 0x34
    ART_JAPAN_CO_LTD = 0x35
    TIANSU_AUTOMATION_CONTROL_SYSTE_CO_LTD = 0x36
    WEINZIERL_ENGINEERING_GMBH = 0x37
    GRUPPO_GIORDANO_IDEA_SPA = 0x38
    ALPHAEOS_AG = 0x39
    TAG_TECHNOLOGIES = 0x3A
    WATTSTOPPER = 0x3B
    PRESSAC_COMMUNICATIONS_LTD = 0x3C
    GIGA_CONCEPT = 0x3E
    SENSORTEC = 0x3F
    JAEGER_DIREKT = 0x40
    AIR_SYSTEM_COMPONENTS_INC = 0x41
    ERMINE_CORP = 0x042
    SODA_GMBH = 0x043
    EKE_AUTOMATION = 0x044
    HOLTER_REGELARMUTREN = 0x045
    ID_RF = 0x046
    DEUTA_CONTROLS_GMBH = 0x047
    EWATTCH = 0x048
    MICROPELT = 0x049
    CALEFFI_SPA = 0x04A
    DIGITAL_CONCEPTS = 0x04B
    EMERSON_CLIMATE_TECHNOLOGIES = 0x04C
    ADEE_ELECTRONIC = 0x04D
    ALTECON = 0x04E
    NANJING_PUTIAN_TELECOMMUNICATIONS = 0x04F
    TERRALUX = 0x050
    MENRED = 0x051
    IEXERGY_GMBH = 0x052
    OVENTROP_GMBH = 0x053
    BUILDING_AUTOMATION_PRODUCTS_INC = 0x054
    FUNCTIONAL_DEVICES_INC = 0x055
    OGGA = 0x056
    ITHO_DAALDEROP = 0x057
    RESOL = 0x058
    ADVANCED_DEVICES = 0x059
    AUTANI_LCC = 0x05A
    DR_RIEDEL_GMBH = 0x05B
    HOPPE_HOLDING_AG = 0x05C
    SIEGENIA_AUBI_KG = 0x05D
    ADEO_SERVICES = 0x05E
    EIMSIG_EFP_GMBH = 0x05F
    VIMAR_SPA = 0x060
    GLEN_DIMLAX_GMBH = 0x061
    PMDM_GMBH = 0x062
    HUBBEL_LIGHTNING = 0x063
    DEBFLEX = 0x64
    PERIFACTORY_SENSORSYSTEMS = 0x65
    WATTY_CORP = 0x66
    WAGO_KONTAKTTECHNIK = 0x67
    KESSEL = 0x68
    AUG_WINKHAUS = 0x69
    DECELECT = 0x6A
    MST_INDUSTRIES = 0x6B
    BECKER_ANTRIEBE = 0x6C
    NEXELEC = 0x6D
    WIELAND_ELECTRIC = 0x6E
    AVIDSEN = 0x6F
    CWS_BOCO_INTERNATIONAL = 0x70
    ROTO_FRANK = 0x71
    ALM_CONTORLS = 0x072
    TOMMASO_TECHNOLOGIES = 0x073
    REHAU = 0x074
    MULTI_USER_MANUFACTURER = 0x7FF
